package edu.ucsd.act.apolrecruitfeed;

import java.io.File;
import java.io.FileFilter;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.InputStream;
import java.io.PrintWriter;
import java.sql.Connection;
import java.sql.DriverManager;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;
import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Date;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Properties;
import java.util.Scanner;

import org.apache.commons.exec.CommandLine;
import org.apache.commons.exec.DefaultExecutor;
import org.apache.commons.io.IOUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;


/**
 * The main class of the APOL UCRecruit data feed batch application. Not to 
 * confuse, but this the Main application class but also contains a standard
 * Java main() method.
 * 
 * This application is triggered in a Unix/Linux environment with a shell 
 * script. See srs/main/bin/recruitfeed.sh.
 * 
 * This class in its run() method goes through the sequential workflow steps:
 * - Extract data from the data sources and generate csv files.
 * - Transmit the files to several UCI servers.
 * 
 * Much of this application relies on properties files and dynamically builds
 * properties keys. This may be confusing at first glance but its purpose is
 * to facilitate this same code to run on UCSD dev, QA, and prod batch servers.
 * 
 * @author David Little
 */
public class Main {

    private static final Logger LOG = LoggerFactory.getLogger(Main.class);
    
    private static final String APP_PROPS = System.getProperty("config.dir") 
            + File.separator + "application.properties";
    
    // since file scp'd on Linux, use Linux new line 
    private static final String NEWLINE = "\n";    
    
    private String batchEnvironment;
    private Properties props;
    private DateFormat timestampFormat;
    
    // maps file types to full file names
    private Map<String, String> extractFilenames;
    
    /**
     * Load properties, register the database driver, and instantiate
     * selected member variables.
     */
    public Main() {
        batchEnvironment = System.getProperty("batch.env");
        props = new Properties();
        try {
            props.load(new FileReader(APP_PROPS));
        } catch (IOException e) {
            LOG.error("Unable to read properties file.", e);
        }
        try {
            Class.forName(props.getProperty("db.driver")); // register driver
        } catch (ClassNotFoundException e) {
            LOG.error("Unable to read properties file.", e);
        }
        timestampFormat = new SimpleDateFormat(props.getProperty("timestamp.format"));
        extractFilenames = new HashMap<String, String>();
    }
    
    /**
     * Extract schools and divisions data from given database connection, 
     * compose a csv file, and save it.
     * 
     * @param conn the database connection
     */
    private void generateSchoolsFile(Connection conn) {
        String outFilename = null;
        Statement stmt = null;
        PrintWriter out = null;
        int outLines = 0;
        try {
            outFilename = composeOutputFilePath(
                    props.getProperty("filename.schools"));
            out = new PrintWriter(new FileWriter(outFilename));
            
            // no output file header line (column names) for this file 

            int divId = 0;
            String divName = null;
            String divAbbr = null;
            String divAddress = null;
            
            String sql = 
                    "select DIV_ID, DIV_NAME, DIV_ABBR, DIV_ADDRESS "
                    + "from APOL_DB.L_DIVISIONS";
            stmt = conn.createStatement();
            ResultSet rs = stmt.executeQuery(sql);
            while (rs.next()) {
                divId = rs.getInt(1);
                divName = rs.getString(2);
                divAbbr = rs.getString(3);
                divAddress = rs.getString(4);
                
                LOG.debug("Division data: {}, {}, {}, {}.", 
                        new Object[] {divId, divName, divAbbr, divAddress});
                
                out.print(quote(String.valueOf(divId)) + "," 
                        + quote(divName) + "," 
                        + quote(divAbbr) + "," 
                        + quote(divAddress) + NEWLINE);
                outLines++;
            }
        } catch (Exception e) {
            LOG.error("Problem generating schools file.", e);
        } finally {
            close(stmt);
            IOUtils.closeQuietly(out);
        }
        extractFilenames.put("schools", outFilename);
        
        LOG.info("Schools file lines: {}.", outLines);
    }
    
    /**
     * Extract department data from given database connection, compose a csv 
     * file, and save it.
     * 
     * @param conn the database connection
     */
    private void generateDepartmentsFile(Connection conn) {
        String outFilename = null;
        Statement stmt = null;
        PrintWriter out = null;
        int outLines = 0;
        try {
            outFilename = composeOutputFilePath(
                    props.getProperty("filename.departments"));
            out = new PrintWriter(new FileWriter(outFilename));
            
            // no output file header line (column names) for this file 

            int deptId = 0;
            String deptName = null;
            int divId = 0;
            
            String sql = 
                    "select DEPT_ID, PPS_DEPT_NAME, DIVISION_ID "
                    + "from APOL_DB.L_DEPT_RECRUIT";
            stmt = conn.createStatement();
            ResultSet rs = stmt.executeQuery(sql);
            while (rs.next()) {
                deptId = rs.getInt(1);
                deptName = rs.getString(2);
                divId = rs.getInt(3);
                
                LOG.debug("Department data: {}, {}, {}.", 
                        new Object[] {deptId, deptName, divId});
                
                out.print(quote(String.valueOf(deptId)) + "," 
                        + quote(deptName) + "," 
                        + quote(String.valueOf(divId)) + NEWLINE);
                outLines++;
            }
        } catch (Exception e) {
            LOG.error("Problem generating departments file.", e);
        } finally {
            close(stmt);
            IOUtils.closeQuietly(out);
        }
        extractFilenames.put("departments", outFilename);

        LOG.info("Departments file lines: {}.", outLines);
    }
    
    /**
     * Extract user and demographic data from given database connection,
     * compose a csv file, and save it.
     * 
     * @param conn the database connection
     */
    private void generateUsersAndDemographicsFiles(Connection conn) {
        String usersFilename = null;
        String demographicsFilename = null;
        Statement stmt = null;
        PrintWriter outUsers = null;
        PrintWriter outDemographics = null;
        int outUsersLines = 0;
        int outDemographicsLines = 0;
        try {
            CodeTranslator codeTrans = new CodeTranslator();
            usersFilename = composeOutputFilePath(
                    props.getProperty("filename.users"));
            demographicsFilename = composeOutputFilePath(
                    props.getProperty("filename.demographics"));
            outUsers = new PrintWriter(new FileWriter(usersFilename));
            outDemographics = new PrintWriter(new FileWriter(demographicsFilename));
            
            // no output file header line (column names) for users file 
            // output file header line (column names) for demographics file
            outDemographics.print("\"external_user_id\",\"ethnicity\",\"gender\"" + NEWLINE);

            String externalId = null;
            String alias = null;
            String firstName = null;
            String lastName = null;
            String lastFirstName = null;
            String email = null;
            String deptName = null;
            int personId = 0;
            String ethnicityCode = null;
            String genderCode = null;
            
            String sql = 
                    "select "
                        + "sys.ID, "                    // 1 external_user_id
                                                        // 2 alias [= email]
                        + "emp.EMP_FIRST_NAME, "        // 3 name
                        + "emp.EMP_LAST_NAME, "         //   3 name
                        + "emp.EMB_EMPLOYEE_NAME, "     // 4 last_first_name
                        + "asa.VANITY_EMAIL_LONG, "     // 5 email_address
                        + "dep.DEPARTMENT_NAME, "       // 6 department
                                                        // 7 department_code [empty]
                                                        // 8 student_id [empty]
                        + "emp.EMB_PERSON_ID, "         // 9 employee_id ["yes"]
                        + "pse.SEC_ETHNICITY_CODE, "    // demographics
                        + "pse.SEC_GENDER_CODE "        // demographics
                    + "from AFFILIATES_DW.AFFILIATES_SAFE_ATTRIBUTES asa "
                    + "join EMPLOYEE.P_EMPLOYEE emp on emp.EMB_PERSON_ID = asa.EMB_PERSON_ID "
                    + "join EMPLOYEE.P_SECURED_EMPLOYEE pse on pse.EMB_PERSON_ID = asa.EMB_PERSON_ID "
                    + "join AFFILIATES_DW.SYSTEM sys on sys.AID = asa.AID "
                    + "join EMPLOYEE.DEPARTMENT dep on dep.DEPARTMENT_CODE = asa.HOME_DEPT "
                    + "where emp.EMP_EMPLOYMENT_STATUS_CODE != 'S' "
                    + "and sys.SYSTEM_ID = '30' "
                    + "and asa.VANITY_EMAIL_LONG is not null";
            stmt = conn.createStatement();
            ResultSet rs = stmt.executeQuery(sql);
            while (rs.next()) {
                externalId = rs.getString(1);
                firstName = rs.getString(2);
                lastName = rs.getString(3);
                lastFirstName = rs.getString(4);
                email = rs.getString(5);
                deptName = rs.getString(6);
                personId = rs.getInt(7);
                ethnicityCode = rs.getString(8);
                genderCode = rs.getString(9);
                
                externalId += "@ucsd.edu";  // suffix domain
                alias = email;              // for UCSD alias is full email
                
                if (email == null) {
                    LOG.info("Null email for person_id {}.", personId);
                } else {
                    if (email.length() == 0) {
                        LOG.info("Empty email for person_id {}.", personId);
                    }
                }

                // write line to users file
                outUsers.print(quote(externalId) + ","                      // 1
                        + quote(alias) + ","                                // 2
                        + quote(firstName.trim() + ' ' + lastName) + ","    // 3
                        + quote(lastFirstName) + ","                        // 4
                        + quote(email) + ","                                // 5
                        + quote(deptName) + ","                             // 6
                        + ","               // dept code (not req'd fr UCSD    7
                        + ","               // empty: not student              8
                        + (personId == 0 ? "" : quote("yes")) + NEWLINE);   // 9
                outUsersLines++;

                // write line to demographics file
                outDemographics.print(quote(externalId) + "," 
                        + quote(codeTrans.translateEthnicity(ethnicityCode)) + "," 
                        + quote(codeTrans.translateGender(genderCode)) + NEWLINE);
                outDemographicsLines++;
            }
        } catch (Exception e) {
            LOG.error("Problem generating users file.", e);
        } finally {
            close(stmt);
            IOUtils.closeQuietly(outUsers);
            IOUtils.closeQuietly(outDemographics);
        }
        extractFilenames.put("users", usersFilename);
        extractFilenames.put("demographics", demographicsFilename);
        
        LOG.info("Users file lines: {}.", outUsersLines);
        LOG.info("Demographics file lines: {}.", outDemographicsLines);
    }
    
    /**
     * Compose a full time-stamped file name and path.
     * 
     * @param filename
     * @return
     */
    protected String composeOutputFilePath(String filename) {
        return System.getProperty("extracts.dir") + File.separator 
                + filename
                + timestampFormat.format(new Date())
                + props.getProperty("filename.extension");
    }
    
    /**
     * Surround a string with double quotes and properly deal with null
     * values, empty strings, and extraneous whitespace.
     * 
     * @param s
     * @return
     */
    protected static String quote(String s) {
        if (s == null) {
            return "";
        }
        String sTrimmed = s.trim();
        return sTrimmed.length() == 0 ? "" : "\"" + sTrimmed + "\"";
    }
    
    /**
     * Convert a delimited string to a list of strings. Used to convert a
     * delimited property value to a list of values
     * 
     * @param delimitedList
     * @param delimiter
     * @return a list of strings
     */
    protected static List<String> convertDelimitedString(
            String delimitedList, String delimiter) {
        List<String> elements = new ArrayList<String>();
        String[] segments = delimitedList.split(delimiter);
        for (String segment : segments) {
            segment = segment.trim();
            if (segment != null && !segment.isEmpty()) {
                elements.add(segment);
            }
        }
        return elements;
    }

    /**
     * Simplify closing a resource without having to handle a checked
     * exception which is usually ignored.
     * 
     * @param conn
     */
    private static void close(Connection conn) {
        if (conn != null) {
            try {
                conn.close();
            } catch (SQLException ignore) {
            }
        }
    }
    
    /**
     * Simplify closing a resource without having to handle a checked
     * exception which is usually ignored.
     * 
     * @param conn
     */
    private static void close(Statement statement) {
        if (statement != null) {
            try {
                statement.close();
            } catch (SQLException ignore) {
            }
        }
    }
    
    /**
     * Transmit via Unix/Linux scp one file to one remote destination
     * 
     * @param fileType which will yield a local file name
     * @param destination which will yield a remote URI
     */
    private void transferFile(String fileType, String destination) {
        String fullFilenameToSend = extractFilenames.get(fileType);
        String fullDestination = props.getProperty("destination." + destination)
                + '/' + props.getProperty("filename." + fileType)
                + props.getProperty("filename.extension");
        try {
            ProcessBuilder pb = new ProcessBuilder("scp", 
                    fullFilenameToSend, fullDestination);
            Process process = pb.start();
            process.waitFor();

            LOG.info("Transferred file {} to {}.", 
                    new File(fullFilenameToSend).getName(), fullDestination);
        } catch (Exception e) {
            String message = String.format("Transfer failed for file %s to %s.", 
                    fullFilenameToSend, fullDestination);
            LOG.error(message, e);
        }
    }
    
    /**
     * Transmit a list of files to a list of destinations.
     */
    private void transferFiles() {
        List<String> fileTypes = convertDelimitedString(
                props.getProperty("transfer.filetypes"), ",");
        List<String> destinations = convertDelimitedString(
                props.getProperty("transfer.destinations"), ",");
        
        // first, file x to each destination, then file y to each ...
        for (String fileType : fileTypes) {
            for (String destination : destinations) {
                transferFile(/*scpTask, */fileType, destination);
            }
        }
    }
    
    /**
     * Deletes extract files that predate the retention period.
     */
    private void deleteOldExtractFiles() {
        long retentionDays = 10;
        try {
            retentionDays = Long.parseLong(props.getProperty("retention.days"));
        } catch (NumberFormatException e) {
            LOG.warn("retention.days in application.properties is not an integer. "
                    + "Default of {} days is being used.", retentionDays);
        }
        final long retentionMillis = retentionDays  // da
                * 24        // hr/da
                * 60        // min/hr
                * 60        // sec/min
                * 1000;     // millis/sec
        final long now = new Date().getTime();
        File extractsFolder = new File(System.getProperty("extracts.dir"));
        File[] oldExtractFiles = extractsFolder.listFiles(new FileFilter() {
            public boolean accept(File file) {
                
                // select files which were written earlier than retention period
                if (file.lastModified() < now - retentionMillis) {
                    return true;
                }

                // do not select files within retention period
                return false;
            }
        });
        for (File oldExtractFile : oldExtractFiles) {
            oldExtractFile.delete();
            
            LOG.info("Deleted old file: {}.", oldExtractFile.getName());
        }
    }
    
    /**
     * Executes the overall workflow. Properties keys are dynamically built
     * based on the batch environment (dev, QA, prod). Different database 
     * connections are required because different source databases are used.
     */
    public void run() {   
        LOG.info("Batch environment: {}.", batchEnvironment);
        
        String keyDir = props.getProperty(batchEnvironment + ".key.dir");
        String pwdDir = props.getProperty(batchEnvironment + ".pwd.dir");
        
        Connection conn = null;
        
        // extract and generate schools (divisions) and departments files
        try {
            File keyFile = new File(keyDir, 
                    props.getProperty(batchEnvironment + ".emlnkdb.key.file"));
            File pwdFile = new File(pwdDir, 
                    props.getProperty(batchEnvironment + ".emlnkdb.pwd.file"));
            Properties emlnkProps = new Properties();
            emlnkProps.load(new FileReader(pwdFile));
            conn = DriverManager.getConnection(
                    props.getProperty("prod.emlnkdb.url"),
                    emlnkProps.getProperty("user"), 
                    Decrypt.decrypt(emlnkProps.getProperty("pwd"), keyFile));
            generateSchoolsFile(conn);
            generateDepartmentsFile(conn);
        } catch (Exception e) {
            LOG.error("Cannot get APOL_DB connection.", e);
            return;
        } finally {
            close(conn);
        }
       
        // extract and generate users and demographics files
        try {
            File keyFile = new File(keyDir, 
                    props.getProperty(batchEnvironment + ".dwdb.key.file"));
            File pwdFile = new File(pwdDir, 
                    props.getProperty(batchEnvironment + ".dwdb.pwd.file"));
            Properties dwProps = new Properties();
            dwProps.load(new FileReader(pwdFile));
            conn = DriverManager.getConnection(
                    props.getProperty("prod.dwdb.url"),
                    dwProps.getProperty("user"), 
                    Decrypt.decrypt(dwProps.getProperty("pwd"), keyFile));
            generateUsersAndDemographicsFiles(conn);
        } catch (Exception e) {
            LOG.error("Cannot get DW_DB connection.", e);
            return;
        } finally {
            close(conn);
        }  

        // scp all files to all destinations
        try {
            transferFiles();
        } catch (Exception e) {
            LOG.error("Problem transmitting files.", e);
            return;
        }
        
        // cleanup old extract files
        deleteOldExtractFiles();
        
        LOG.info("Finished.");
    }
    
    /**
     * Simply pass control to the run() method.
     * 
     * @param args not used
     */
    public static void main(String[] args) {
        Main app = new Main();
        app.run();
    }

}
